/*
  Copyright 2006,2008 CNRS

  Authors Florent Lamiraux (CNRS) and Mathieu Poirier (CNRS)

*/




#ifndef HPP_WALK_PLANNER_H
#define HPP_WALK_PLANNER_H

/*************************************
INCLUDE
**************************************/

#include <iostream>

#include "hppWalkFootPlanner/hppWalkFootPlanner.h"
#include "KineoWorks2/kwsPathOptimizer.h"
#include "KineoWorks2/kwsRoadmapBuilder.h"
#include "KineoWorks2/kwsConfig.h"

/*************************************
STRUCTURES
**************************************/
/*************************************
CLASS
**************************************/

class ChppWalkPatGen;

/**
   \brief Implement a two step walk planning strategy.

   The motion planning strategy is based on an iterative method that plan successive motions. In a first step, a path is computed for the bounding box of the robot and then this path is animated using JRL walk pattern generator. This iterative strategy uses two instanciations of the humanoid robot:
   \li a bounding box and
   \li a fully actuated humanoid robot.

   <b> How to use this class</b>

   To use this class, 
   \li create an object ChppWalkPlanner, by specifying the step size,
   \li insert a humanoid robot using ChppPlanner::addHppProblem in the vector of problems, 
   \li put this robot in the starting configuration of a walk motion,
   \li call ChppWalkPlanner::initializeProblem to build the bounding box of the humanoid robot,
   \li add obstacles using ChppPlanner::addObstacle 
   \li choose initial and goal configurations of the box using ChppWalkPlanner::initConfig and ChppWalkPlanner::goalConfig and 
   \li call ChppWalkPlanner::solve to compute a collision-free walking motion for the humanoid robot.
  
   To get the resulting path, use ChppPlanner::getPath with problemId=1 and pathId = ChppPlanner::getNbPaths(1)

   Note that the default steering method for the bounding box is ChppWalkSteeringMethod, but it can be changed using ChppPlanner API.
*/
class ChppWalkPlanner : public ChppWalkFootPlanner {
public: 
  /**
     \brief Constructor

     \param inStepLength Abscissa of the landing foot in the frame of the support foot in straight line motions.
     \param inStepSpace Absolute value of the ordinate of the landing foot in the frame of the support foot
     \param inSecurityDistance Distance above which centers of feet should remain in two consecutive steps.
     \param inBoxLateralTolerance lateral growing of bounding box to include swinging motion during walk.
  */
  ChppWalkPlanner(double inStepLength, double inStepSpace, 
		  double inSecurityDistance,
		  double inBoxLateralTolerance);

  /**
     \brief Destructor
  */
  virtual ~ChppWalkPlanner();

  /**
     \name Problem definition
     @{
  */
  
  /**
     \brief Set initial configuration of Walk Planning problem
     \param x position and orientation of the robot.
     \param y position and orientation of the robot.
     \param theta position and orientation of the robot.

     Call ChppFootPlanner::initConfig() and set humanoid robot in corresponding
     half-sitting configuration;
  */
  ktStatus initConfig(double x, double y, double theta);
  
  /**
     \brief Set goal configuration of Walk Planning problem
     \param x position and orientation of the robot.
     \param y position and orientation of the robot.
     \param theta position and orientation of the robot.

     Call ChppFootPlanner::goalConfig() and set humanoid robot in corresponding
     half-sitting configuration;
  */
  ktStatus goalConfig(double x, double y, double theta);

  /**
     \brief Initialize a Walk Planning problem.
     Compute the bounding box of the first robot in problem  vector (in given configuration) and create a robot with this box. Then insert this new robot at beginning of problem vector.
     
     The box robot is placed in the first problem by a push_front().

  */
  virtual ktStatus initializeProblem();
  
  /** 
      @}
  */


  /**
     \name Problem resolution
     @{
  */
  /**
     \brief Iteratively solve a walk motion planning problem for a humanoid robot. \     
     \return KD_OK or KD_ERROR
  */
  virtual ktStatus solve();

  /**
     \brief Get the all the element to compute the solution on robot. 
     
     \param  outRobotPath the hpp path of humanoid robot
     \param  outTimeRef the Time Reference for all the element
     \param  outZMPRef  a vector of ZMPRef(3) ... vector resize on TimeRef size
     \param  outWaistPosition a vector of WaistPosition(6) ... vector resize on TimeRef size
  
     \return KD_OK or KD_ERROR
  */
  ktStatus solution(std::vector<CkwsConfigShPtr> &outRobotPath, 
		    std::vector<double> &outTimeRef, 
		    std::vector<std::vector<double> > &outZMPRef, 
		    std::vector<std::vector<double> > &outWaistPosition );

  /**
     \brief compute the 2st phase : compute full trajectory of humanoid robot
     \return KD_OK or KD_ERROR
  */
  ktStatus computeHumanoidTrajectory() ;


  /** 
   *@}
   */

private:

  /**
     \brief Reset Trajectory of humanoid robot

  */
  void resetHumanoidTrajectory() {
    attRobotPath.clear();
    attTimeRef.clear();
    attZMPRef.clear();
    attWaistPosition.clear();

  };

  /// Full path of humanoid robot
  std::vector<CkwsConfigShPtr> attRobotPath ;
  
  /// Time reference along trajectory
  std::vector<double> attTimeRef ;
  
  /// ZMP reference of along trajectory
  std::vector<std::vector<double> > attZMPRef ;
 
  /// Waist position along trajectory
  std::vector<std::vector<double> > attWaistPosition ;

  /// \brief Object that compute the stack of steps and the robot walk trajectory.
  ChppWalkPatGen* attHppPatGen;

  /**
     \brief Lateral tolerance of the bounding box

     The bounding box of the robot is enlarged by this value.
  */
  double attBoxLateralTolerance;

};


#endif


