/*
  Copyright 2006,2008 CNRS

  Authors: Mathieu Poirier (CNRS) and Florent Lamiraux (CNRS)

*/

#ifndef HPP_PAT_GEN_H
#define HPP_PAT_GEN_H

/*************************************
INCLUDE
**************************************/

#include <iostream>
#include <fstream>


#include <MatrixAbstractLayer/MatrixAbstractLayer.h>
#include <walkGenJrl/PatternGeneratorInterface.h>
#include "KineoWorks2/kwsPath.h"
#include "hppModel/hppDevice.h"
#include "hppWalkFootPlanner/hppWalkStepSequence.h"

using namespace std ;
using namespace PatternGeneratorJRL;

KIT_PREDEF_CLASS(CkwsPlusDirectPath);

/*************************************
STRUCTURES
**************************************/

/*************************************
CLASS
**************************************/
/**
   \brief  API to use the walkGenJRL library for collision free walk planning
*/
class ChppWalkPatGen
{

public:

  /**
     \brief  constructor that sets the robot in half-sitting position.

     \param inStepLength Abscissa of the landing foot in the frame of the support foot in straight line motions.
     \param inStepSpace Absolute value of the ordinate of the landing foot in the frame of the support foot
  */
  ChppWalkPatGen(double inStepLength, double inStepSpace);

  /**
     \brief  destructor
  */
  ~ChppWalkPatGen() ;

  /**
     \brief Read only access to step spacing
  */
  double stepSpacing() const {
    return attStepSpacing;
  };

  /**
     \brief Reset the configuration of the robot in half-sitting position.
  */
  void reset();

  /**
     \brief  Initialize a PatternGenerator (interface) with several variables

     \li omega  : 0.0
     \li stepheight : 0.07
     \li singlesupporttime : 0.78
     \li doublesupporttime : 0.02
     \li armparameters :  0.5
     \li LimitsFeasibility : 0.0
     \li ZMPShiftParameters :  0.015 0.015 0.015 0.015
     \li TimeDistributionParameters : 2.0 3.7 1.7 3.0
     \li UpperBodyMotionParameters :  -0.1 -1.0 0.0

  */
  void initWalkPatGen(CjrlHumanoidDynamicRobot* inRobot);

  /**
     \brief  compute all the Q configuration in 5 ms time for OpenHRP (in FILE .pos, ...) and KINEOWORKS (in structure pathHrp2)
     \param hppDevice kwsDevice of HRP2
     \param outPathHumanoid  kwsPath of HRP2 to be play in the kppInterface
     \param outDiscretizedPath vector of HRP2 kwsConfigs equal to HRP2 kwsPath for external use as OpenHrp
     \param outHRP2TimeRef vector of Time reference for outHRP2Path outHRP2ZMPRef and outHRP2WaistPosition
     \param outHRP2ZMPRef  vector or ZMP trajectory corresponding to the outHRP2Path
     \param outHRP2WaistPosition vector or Waist trajectory (position and orientation) corresponding to the outHRP2Path
     \param inRelativeFootPrintVector step sequence

     The trajectory of the robot is moved by (ChppWalkPatGen::m_xInit
     ChppWalkPatGen::m_yInit ChppWalkPatGen::m_thetaInit). See
     ChppWalkPatGen::computeAStepSequenceFromAPath for explanation.
  */
  ktStatus computeAllQFromAStepSequence( ChppDeviceShPtr hppDevice, CkwsPathShPtr& outPathHumanoid,
					 std::vector<CkwsConfigShPtr> &outDiscretizedPath, std::vector<double> &outHRP2TimeRef,
					 std::vector<std::vector<double> > &outHRP2ZMPRef, std::vector<std::vector<double> > &outHRP2WaistPosition,
					 const ChppWalkStepSequence& inRelativeFootPrintVector);

  /**
     \brief  compute a stream for Any Curve : step seq : ................. from a step Sequence for PatternGeneratorInterface
     \param inRelativeFootPrintVector step sequence.
  */
  void computeStreamFromStepSequence(const ChppWalkStepSequence& inRelativeFootPrintVector);

private :

  /**
     \brief Length of the steps
     Abscissa of the landing foot in the frame of the support foot in straight line motions.
  */
  double attStepLength;

  /**
     \brief Space between feet
     Absolute value of the ordinate of the landing foot in the frame of the support foot
  */
  double attStepSpacing;

  /**
     \brief String stream passed to pattern generator interface
  */
  std::istringstream attStringStream;

  /**
     \brief Interface with walk patter generator
  */
  PatternGeneratorInterface* attWalkPatGen;
};



#endif


